//
// binarysearch.C
//
// Author:               Sameer Nene
// Date:                 05/26/94
// Version:              1.0
// Modification History:
//   06/26/94: Changes necessary because of change in interface.
//   05/02/95: Major changes in search(). Added code to select smallest range
//             first.
//   06/29/96: Made class non-persistent. Also removed dependence on DataSet.
//
// Bugs:
//
// Classes:
//   BinarySearch
//
// Notes:
//   This module contains implementation of classes declared in binarysearch.h
//
// ----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include <values.h>
#include "errorscope.h"
#include "vector.h"
#include "binarysearch.h"

int BinarySearch::d_pos;
double *BinarySearch::d_sa_p;
Vector BinarySearch::d_pv;				       
				       
int BinarySearch::compare(const void *e1, const void *e2)
{
  return d_sa_p[*((int *)e1)] - d_sa_p[*((int *)e2)] < 0 ? -1 : 1;
}

int BinarySearch::bsearchL(int col, double d) const
{
  Vector &v = d_data[col];
  int b = 0, c, t = v.length() - 1;
  double q;
  
  while(t - b > 1) {
    c = (b + t) >> 1;
    q = v.element(c);
    if(d < q)
      t = c;
    else if(d > q)
      b = c;
    else
      return c;
  }

  return ((d <= v.element(b)) ? b : t);
}

int BinarySearch::bsearchR(int col, double d) const
{
  Vector &v = d_data[col];
  int b = 0, c, t = v.length() - 1;
  double q;
  
  while(t - b > 1) {
    c = (b + t) >> 1;
    q = v.element(c);
    if(d < q)
      t = c;
    else if(d > q)
      b = c;
    else
      return c;
  }

  return ((d >= v.element(t)) ? t : b);
}

BinarySearch::BinarySearch() : d_bmaps_p(0), d_fmaps_p(0), d_thresh(.1)
{
}

BinarySearch::BinarySearch(const VectAry &va) : d_data(va[0].length()), d_thresh(.1)
{
  int i, j, k, d_i, *p1, *p2;
  
  d_data[0].length(j = va.getSize());
  d_bmaps_p = new int*[k = va[0].length()];
  d_fmaps_p = new int*[k];
  d_sa_p = new double[j];

  for(d_i = 0; d_i < k; ++d_i) {
    p1 = d_bmaps_p[d_i] = new int[j];
    d_data[d_i].length(j);
    for(i = 0; i < j; ++i) {
      p1[i] = i;
      d_sa_p[i] = va[i].element(d_i);
    }
    qsort(p1, j, sizeof(int), compare);
    for(i = 0; i < j; ++i)
      d_data[d_i].element(i, d_sa_p[p1[i]]);
    p2 = d_fmaps_p[d_i] = new int[j];
    for(i = 0; i < j; ++i)
      p2[p1[i]] = i;
  }
}

BinarySearch::~BinarySearch()
{
  for(int i = 0, j = d_data.getSize(); i < j; ++i) {
    delete[] d_bmaps_p[i];
    delete[] d_fmaps_p[i];
  }
  delete[] d_bmaps_p;
  delete[] d_fmaps_p;
  delete[] d_sa_p;
}

int BinarySearch::search(const Vector &v) const
{
  int *b, *d, i, j, k, l, n, o, *t, *p, u, bot, top, *bmap, *list;
  double m, q, tsq;
  Vector r(d_n.length());

  b = new int[j = d_data.getSize()];
  t = new int[j];
  for(i = 0, k = d_data[0].length() - 1, l = 0; i < j; ++i) {
    b[i] = bsearchL(i, v.element(i) - d_thresh);
    t[i] = bsearchR(i, v.element(i) + d_thresh);
  }

  d = new int[j];
  for(i = 0; i < j; ++i)
    d[i] = i;
  for(i = 0, n = 1; i < j - 1 && n != 0; ++i)
    for(k = n = 0; k < j - i - 1; ++k)
      if(t[d[k + 1]] - b[d[k + 1]] < t[d[k]] - b[d[k]]) {
	l = d[k];
	d[k] = d[k + 1];
	d[k + 1] = l;
	n = 1;
      }

  bmap = d_bmaps_p[d[0]];
  p = d_fmaps_p[d[1]];
  list = new int[t[d[0]] - b[d[0]] + 1];
  
  for(i = b[d[0]], k = t[d[0]], n = 0, bot = b[d[1]], top = t[d[1]]; i <= k; ++i) {
    o = p[l = bmap[i]]; 
    if(o >= bot && o <= top)
      list[n++] = l;
  }

  for(i = 2, j = d_data.getSize(); i < j; ++i) {
    p = d_fmaps_p[d[i]];
    bot = b[d[i]];
    top = t[d[i]];
    for(k = 0, l = n, n = 0; k < l; ++k) {
      o = p[u = list[k]]; 
      if(o >= bot && o <= top)
	list[n++] = u;
    }
  }

  delete[] b;
  delete[] t;
  delete[] d;

  Vector s(j);
  
  m = MAXDOUBLE;
  for(k = 0, l = n, tsq = d_thresh * d_thresh; k < l; ++k) {
    o = list[k];
    for(i = 0; i < j; ++i)
      s.element(i, d_data[i].element(d_fmaps_p[i][o]));
    if((q = s.distance(v)) < m)
      if(q <= tsq) {
	m = q;
	n = list[k];
      }
  }
  
  delete[] list;

  return n;
}

void BinarySearch::setSearchParameters(const Vector &v)
{
  d_thresh = v.element(0);
}

ErrorScope::Error BinarySearch::get(FILE *file)
{
  int i, j, k;
  ErrorScope::Error error;

  j = d_data.getSize();

  if((error = d_data.get(file)) != OK)
    return error;

  for(i = 0; i < j; ++i) {
    delete[] d_bmaps_p[i];
    delete[] d_fmaps_p[i];
  }
  delete[] d_bmaps_p;
  delete[] d_fmaps_p;
  d_bmaps_p = new int*[j = d_data.getSize()];
  d_fmaps_p = new int*[j = d_data.getSize()];
  k = d_data[0].length();

  for(i = 0; i < j; ++i) {
    d_bmaps_p[i] = new int[k];
    if(fread((char *)d_bmaps_p[i], sizeof(int), k, file) != k)
      return FILE_READ_ERROR;
  }
  
  for(i = 0; i < j; ++i) {
    d_fmaps_p[i] = new int[k];
    if(fread((char *)d_fmaps_p[i], sizeof(int), k, file) != k)
      return FILE_READ_ERROR;
  }
  
  return OK;
}

ErrorScope::Error BinarySearch::put(FILE *file) const
{
  int i, j, k;
  ErrorScope::Error error;
  
  if((error = d_data.put(file)) != OK)
    return error;

  k = d_data[0].length();

  for(i = 0, j = d_data.getSize(); i < j; ++i)
    if(fwrite((char *)d_bmaps_p[i], sizeof(int), k, file) != k)
      return FILE_WRITE_ERROR;
  
  for(i = 0, j = d_data.getSize(); i < j; ++i)
    if(fwrite((char *)d_fmaps_p[i], sizeof(int), k, file) != k)
      return FILE_WRITE_ERROR;
  
  return OK;
}

const VectAry& BinarySearch::getData() const
{
  return d_data;
}
